

def add_missing_resources(template:dict, src:dict) -> bool:
    dirty = False
    for top_level in ['Parameters', 'Resources', 'Conditions', 'Mappings']:
        for name, value in src.get(top_level, {}).items():
            if not template.get(top_level,{}).get(name):
                if top_level not in template:
                    template[top_level] = {}
                template[top_level][name] = value
                dirty = True
    return dirty


from ruamel.yaml.comments import TaggedScalar


def replace_references(template, old_reference, new_reference) -> bool:
    """
    replaces CloudFormation references { "Ref": old_reference } with { "Ref": new_reference } in `template`.
    returns True if one or more references where made.
    """
    result = False
    if isinstance(template, dict):
        for name, value in template.items():
            if name == 'Ref' and value == old_reference:
                template['Ref'] = new_reference
                result = True
            elif isinstance(value, TaggedScalar) and value.tag and value.tag.value == '!Ref' and value.value == old_reference:
                value.value = new_reference
                result = True
            else:
                result = replace_references(template[name], old_reference, new_reference) or result
    elif isinstance(template, list):
        for i, value in enumerate(template):
            result = replace_references(value, old_reference, new_reference) or result
    return result



"""
update_packer unit test.
"""
import boto3
import unittest
from update_packer.base import UpdatePackerUnitTestBase
from aws_cfn_update.packer_ami_updater import PackerAMIUpdater


class UpdatePackerUnitTest(UpdatePackerUnitTestBase):
    def test_update_packer(self) -> None:
        packer = {
            "builders": [
                {
                    "type": "amazon-ebs",
                    "communicator": "winrm",
                    "region": "eu-central-1",
                    "source_ami_filter": {
                        "filters": {
                            "virtualization-type": "hvm",
                            "name": "Windows_Server-2016-English-Full-Base-2020.01.10",
                            "root-device-type": "ebs",
                        },
                        "owners": ["801119661308"],
                    },
                }
            ]
        }
        old_name = packer["builders"][0]["source_ami_filter"]["filters"]["name"]
        boto3.DEFAULT_SESSION = self.session
        updater = PackerAMIUpdater()
        updater.ami_name_pattern = "Windows_Server-2016-English-Full-Base*"
        updater.packer = packer
        updater.update()
        assert updater.dirty
        new_name = packer["builders"][0]["source_ami_filter"]["filters"]["name"]
        assert old_name != new_name

    def setUp(self):
        super().setUp()

    def tearDown(self):
        super().tearDown()


if __name__ == "__main__":
    unittest.main()



"""
Generated base class for the update_packer unit test. The setup will provide stubbed
responses for all recorded requests.

** generated code - do not edit **
"""
import unittest
import botocore.session
from botocore.stub import Stubber

from update_packer import call_00001_describe_images


class UpdatePackerUnitTestBase(unittest.TestCase):
    def setUp(self) -> None:
        """
        add stubs for all AWS API calls
        """
        self.session = botocore.session.get_session()
        self.clients = {
            service: self.session.create_client(service) for service in ["ec2"]
        }
        self.stubs = {
            service: Stubber(client) for service, client in self.clients.items()
        }

        self.stubs["ec2"].add_response(
            "describe_images",
            call_00001_describe_images.response,
            call_00001_describe_images.request,
        )

        for _, stub in self.stubs.items():
            stub.activate()
        self.session.client = lambda x: self.clients[x]

    def tearDown(self) -> None:
        """
        check all api calls were executed
        """
        for service, stub in self.stubs.items():
            stub.assert_no_pending_responses()
            stub.deactivate()



#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
#   Copyright 2018 binx.io B.V.
import fnmatch
import re
import copy
import click
import json
import boto3
from typing import List
from ruamel.yaml.comments import TaggedScalar, CommentedSeq

from .cfn_updater import CfnUpdater

import logging

def is_reference(obj, fld, name):
    if isinstance(obj, dict):
        if fld == "Ref":
            return obj["Ref"] == name
        elif fld == "FN::GetAtt":
            return len(obj[fld]) > 0 and obj[fld][0] == name
        elif fld == "FN::Sub":
            return isinstance(obj[fld], str) and references_in_sub(obj[fld], name)

def references_in_sub(sub, name):
    return list(filter(lambda n : n == name, map(lambda n: n.split('.')[0], re.findall(r'\${([^!][^}]*)}', sub))))

def is_tag_reference(obj, name):
    if isinstance(obj, TaggedScalar):
        if obj.tag.value == '!Ref':
            return str(obj.value) == name
        elif obj.tag.value == '!Sub':
            return references_in_sub(obj.value, name)
        elif obj.tag.value == '!GetAtt':
            return str(obj.value).split('.')[0] == name
    elif isinstance(obj, CommentedSeq):
        if obj.tag.value == '!GetAtt':
            return len(obj) > 1 and obj[0] == name
    return False

def remove_resource_from_template(template:dict, name:str):
    resources:dict = template.get("Resources")
    if resources and name in resources:
        resources.pop(name)
        remove_all_references(template, name)
        return True
    outputs:dict = template.get("Resources")
    if outputs and name in outputs:
        remove_all_references(template, name)
        return True
    return False

    return False

def remove_all_references(template, name):
    to_remove = {}
    for toplevel in template:
        if isinstance(template[toplevel], dict):
            to_remove[toplevel] = []

    for toplevel in to_remove:
        for obj in template[toplevel]:
            if has_reference(
                    template[toplevel][obj], name, []
            ):
                to_remove[toplevel].append(obj)
    for toplevel in to_remove:
        for obj in to_remove[toplevel]:
            logging.info(
                "Removing object %s from %s, as it references %s",
                obj, toplevel, name
            )
            template[toplevel].pop(obj)

def has_reference(obj, name, path):
    result = False
    if isinstance(obj, dict):
        for fld in obj:
            if is_reference(obj, fld, name):
                logging.info(
                    "INFO: %s to %s found in %s",
                    fld, name, ".".join(path)
                )
                result = True
            else:
                nested_path = path[:]
                nested_path.append(fld)
                result = result or has_reference(
                    obj[fld], name, nested_path
                )
    elif isinstance(obj, TaggedScalar):
        if is_tag_reference(obj, name):
            logging.info(
                "INFO: %s to %s found in %s",
                obj.tag, name, ".".join(path)
            )
            result = True

    elif isinstance(obj, CommentedSeq):
        if is_tag_reference(obj, name):
            logging.info(
                "INFO: %s to %s found in %s",
                obj.tag, name, ".".join(path)
            )
            result = True
        else:
            for i in range(len(obj)):
                nested_path = path[:]
                nested_path.append("[%d]" % i)
                if has_reference(obj[i], name, nested_path):
                    logging.info(
                        "INFO: %s too %s found in array %s",
                        obj[i], name, ".".join(nested_path)
                    )
                    result = True

    elif isinstance(obj, list):
        for i in range(len(obj)):
            nested_path = path[:]
            nested_path.append("[%d]" % i)
            if has_reference(obj[i], name, nested_path):
                logging.info(
                    "INFO: Ref:%s found in array %s",
                    name, ".".join(nested_path)
                )
                result = True

    return result



class ResourceRemover(CfnUpdater):
    """
    Removes the specified CloudFormation resource and all resources that reference it.
    """

    def __init__(self):
        super(ResourceRemover, self).__init__()
        self.resource_name = None

    def update_template(self):
        if remove_resource_from_template(self.template, self.resource_name):
            self.dirty = True


@click.command(name='remove-resource', help=ResourceRemover.__doc__)
@click.option('--resource', required=True, help='to remove from the template')
@click.argument('path', nargs=-1, required=True, type=click.Path(exists=True))
@click.pass_context
def remove_resource(ctx, resource, path):
    updater = ResourceRemover()
    updater.dry_run = ctx.obj['dry_run']
    updater.verbose = ctx.obj['verbose']
    updater.resource_name = resource
    updater.update(path)



#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
#   Copyright 2018 binx.io B.V.
import binascii
import ssl
import sys
from urllib.parse import urlparse

import click
import requests
from cryptography import x509
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes
from ruamel.yaml import CommentedSeq
from aws_cfn_update.cfn_updater import CfnUpdater


class OIDCProviderThumbprintsUpdater(CfnUpdater):
    """
    Updates the thumbprints list of an AWS::IAM::OIDCProvider.

    By default, it updates the thumbprints of all OIDCProviders specified
    templates.
    """

    def __init__(self):
        super(OIDCProviderThumbprintsUpdater, self).__init__()
        self.url = None
        self.verbose = False
        self.dry_run = False
        self.append = False

    def new_value(self, definition):
        if self.with_fn_sub:
            return isinstance(definition, dict) and definition.get(
                "Fn::Sub"
            ) == self.definition, {"Fn::Sub": self.definition}
        else:
            return definition == self.definition, self.definition

    def update_template(self):
        """
        updates the Thumbprints of OIDCProviders
        """
        resources: dict = self.template.get("Resources", {})
        for name in self.all_matching_oidc_providers(resources):
            provider = resources[name]
            self.update_thumbprints(name, provider)

    @staticmethod
    def is_type_oidc_provider(resource):
        """
        returns true if the resource is of type AWS::IAM::OIDCProvider
        """
        return resource.get("Type", "") == "AWS::IAM::OIDCProvider"

    def is_matching_oidc_provider(self, provider):
        """
        Returns true if there is a match on self.url and the url starts with https://.
        """
        url = provider.get("Properties", {}).get("Url")
        if not (url and url.startswith("https://")):
            return False

        if not self.url:
            return True

        return url == self.url

    def all_matching_oidc_providers(self, resources):
        return filter(lambda n: self.is_matching_oidc_provider(resources[n]), resources)

    def main(self, url, append, path, dry_run, verbose):
        self.dry_run = dry_run
        self.verbose = verbose
        self.url = url
        self.append = append
        self.update(path)

    @staticmethod
    def get_public_key(url: str):
        wks = f"{url}/.well-known/openid-configuration"
        response = requests.get(wks, headers={"Accept": "application/json"})
        if response.status_code != 200:
            raise ValueError(
                "expected 200 from %s, got %d, %s",
                url,
                response.status_code,
                response.text,
            )

        configuration = response.json()
        if "jwks_uri" not in configuration:
            raise ValueError("%s did not return a proper openid configuration", wks)

        jwks_uri = urlparse(configuration["jwks_uri"])

        conn = None
        try:
            conn = ssl.create_connection((jwks_uri.netloc, 443))
            context = ssl.SSLContext(ssl.PROTOCOL_SSLv23)
            sock = context.wrap_socket(conn, server_hostname=jwks_uri.netloc)
            certificate = ssl.DER_cert_to_PEM_cert(sock.getpeercert(True))
            return x509.load_pem_x509_certificate(
                certificate.encode("ascii"), default_backend()
            )
        finally:
            if conn:
                conn.close()

    def update_thumbprints(self, name, provider):

        url = provider.get("Properties", {}).get("Url")
        public_key = self.get_public_key(url)
        sha1 = public_key.fingerprint(hashes.SHA1())
        fingerprint = binascii.hexlify(sha1).decode("ascii").lower()

        thumbprints = provider.get("Properties", {}).get("ThumbprintList", CommentedSeq())
        exists = list(filter(lambda f: f.lower() == fingerprint, thumbprints))
        if exists:
            if self.verbose:
                sys.stderr.write(
                    f"INFO: fingerprint of {url} for OIDC provider {name} already in thumbprint list\n"
                )
            return

        if self.verbose:
            subject = public_key.subject.rfc4514_string()
            issuer = public_key.issuer.rfc4514_string()
            sys.stderr.write(
                f"INFO: updating fingerprint of {url} for OIDC provider {name}, {subject} issued by {issuer}\n"
            )

        sys.stderr.write(
            f"INFO: updating fingerprint of {url}, for OIDC provider {name} to {fingerprint}, valid until {public_key.not_valid_after}\n"
        )
        self.dirty = True


        if not isinstance(thumbprints, CommentedSeq):
            new_thumbprints = CommentedSeq()
            if self.append:
                new_thumbprints.extend(thumbprints)
            thumbprints = new_thumbprints

        thumbprints.append(fingerprint)
        thumbprints.yaml_add_eol_comment(f"valid until {public_key.not_valid_after}", len(thumbprints)-1)
        provider["Properties"]["ThumbprintList"] = thumbprints


@click.command(
    name="oidc-provider-thumbprints", help=OIDCProviderThumbprintsUpdater.__doc__
)
@click.option("--url", required=False, type=str, help="of the OIDC provider to update")
@click.option("--append", required=False, is_flag=True, help="append the fingerprint")
@click.argument("path", nargs=-1, required=True, type=click.Path(exists=True))
@click.pass_context
def update_oidc_provider_thumbprint(ctx, url, append, path):
    updater = OIDCProviderThumbprintsUpdater()
    updater.main(url, append, list(path), ctx.obj["dry_run"], ctx.obj["verbose"])



#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
#   Copyright 2018 binx.io B.V.
import click

from .add_missing_resources import add_missing_resources
from .cfn_updater import CfnUpdater, read_template


class AddNewResources(CfnUpdater):
    """
    add resources that exist in the new template and not in the existing template
    """

    def __init__(self):
        super(AddNewResources, self).__init__()
        self.source:dict = {}

    def update_template(self):
        self.dirty = add_missing_resources(self.template, self.source)


@click.command(name='add-new-resources', help=AddNewResources.__doc__)
@click.option('--source', required=True, help='template to add resources from', type=click.Path(exists=True))
@click.argument('path', nargs=1, required=True, type=click.Path(exists=True))
@click.pass_context
def add_new_resources(ctx, source, path):
    updater = AddNewResources()
    updater.dry_run = ctx.obj['dry_run']
    updater.verbose = ctx.obj['verbose']
    updater.source = read_template(source)
    updater.update(path)



#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
#   Copyright 2018 binx.io B.V.
import sys
from .cfn_updater import CfnUpdater
from ruamel.yaml.scalarstring import PreservedScalarString


class LambdaInlineCodeUpdater(CfnUpdater):
    """
    Updates the inline code of an AWS::Lambda::Function resource.

\b
  ELBListenerRuleProvider:
    Type: AWS::Lambda::Function

\b
  ELBListenerRuleProvider:
    Type: AWS::Lambda::Function
    Properties:
      Code:
        ZipFile:
          import boto3
          import cfnresponse
          ELB = boto3.client('elbv2')
          ...
    """

    def __init__(self):
        super(LambdaInlineCodeUpdater, self).__init__()
        self._image = []

    def update_template(self):
        """
        updates the Code property of a AWS::Lambda::Function resource of name `self.resource` to `self.code`
        """
        resource = self.template.get('Resources', {}).get(self.resource, None)
        if resource and resource['Type'] == 'AWS::Lambda::Function':
            code = resource.get('Properties', {}).get('Code', {})
            old_code = code['ZipFile'] if 'ZipFile' in code else None
            if old_code != self.code:
                sys.stderr.write(
                    'INFO: updating inline code of lambda {} in {}\n'.format(self.resource, self.filename))
                if 'Properties' not in resource:
                    resource['Properties'] = {}
                if 'Code' not in resource['Properties']:
                    resource['Properties']['Code'] = {}
                resource['Properties']['Code'] = {'ZipFile': PreservedScalarString(self.code)}
                self.dirty = True
        elif resource:
            sys.stderr.write(
                'WARN: resource {} in {} is not of type AWS::Lambda::Function\n'.format(self.resource, self.filename))

    def main(self, resource, code, paths, dry_run, verbose):
        self.resource = resource
        self.code = code
        self.dry_run = dry_run
        self.verbose = verbose
        self.update(paths)



#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
#   Copyright 2018 binx.io B.V.
import re
from copy import copy
from datetime import datetime
from aws_cfn_update.packer_ami_updater import main as packer_latest_ami

import click
import click_datetime
import pytz

from aws_cfn_update.config_rule_inline_code_updater import ConfigRuleInlineCodeUpdater
from aws_cfn_update.container_image_updater import ContainerImageUpdater
from aws_cfn_update.cron_schedule_expression_updater import CronScheduleExpressionUpdater
from aws_cfn_update.latest_ami_updater import AMIUpdater
from aws_cfn_update.rest_api_body_updater import RestAPIBodyUpdater
from aws_cfn_update.lambda_inline_code_updater import LambdaInlineCodeUpdater
from aws_cfn_update.statemachine_updater import update_state_machine_definition
from aws_cfn_update.remove_resource import remove_resource
from aws_cfn_update.add_new_resources import add_new_resources
from aws_cfn_update.oidc_provider_thumbprints_updater import update_oidc_provider_thumbprint


@click.group()
@click.option('--dry-run', is_flag=True, default=False,
              help='do not change anything, just show what is going to happen')
@click.option('--verbose', is_flag=True, default=False, help='show more output')
@click.pass_context
def cli(ctx, dry_run, verbose):
    """Programmatically update CloudFormation templates"""
    ctx.obj = copy(ctx.params)


def validate_image(ctx, param, value):
    if re.match(r'(?:.+/)?([^:]+)(?::.+)?', value):
        return value
    else:
        raise click.BadParameter('"{}" is not a valid docker image specification'.format(value))


@cli.command(name='container-image', help=ContainerImageUpdater.__doc__)
@click.option('--image', required=True, callback=validate_image, help='to update to')
@click.argument('path', nargs=-1, required=True, type=click.Path(exists=True))
@click.pass_context
def task_image(ctx, image, path):
    updater = ContainerImageUpdater()
    updater.main(image, ctx.obj['dry_run'], ctx.obj['verbose'], list(path))


@cli.command(name='latest-ami', help=AMIUpdater.__doc__)
@click.option('--ami-name-pattern', required=True, help='glob style pattern of the AMI image name to use')
@click.option('--add-new-version', is_flag=True, default=False, help='of the AMI resource and replace all references')
@click.argument('path', nargs=-1, required=True, type=click.Path(exists=True))
@click.pass_context
def ami_image_update(ctx, ami_name_pattern, add_new_version, path):
    updater = AMIUpdater()
    updater.main(ami_name_pattern, ctx.obj['dry_run'], ctx.obj['verbose'], add_new_version, list(path))


@cli.command(name='cron-schedule-expression', help=CronScheduleExpressionUpdater.__doc__)
@click.option('--timezone', required=False, help='to use to calculate the UTC time', default="Europe/Amsterdam")
@click.option('--date', type=click_datetime.Datetime(format='%Y-%m-%d'), default=datetime.now(),
              help='to use as reference date')
@click.argument('path', nargs=-1, required=True, type=click.Path(exists=True))
@click.pass_context
def cron_schedule_expression(ctx, timezone, date, path):
    updater = CronScheduleExpressionUpdater()
    try:
        tz = pytz.timezone(timezone)
        updater.main(tz, date, ctx.obj['dry_run'], ctx.obj['verbose'], list(path))
    except pytz.exceptions.UnknownTimeZoneError:
        raise click.BadParameter('invalid timezone specified', ctx=ctx, param='timezone')


@cli.command(name='rest-api-body', help=RestAPIBodyUpdater.__doc__)
@click.option('--resource', required=True, help='AWS::ApiGateway::RestApi body to update')
@click.option('--open-api-specification', required=True, type=click.Path(exists=True), help='defining the interface')
@click.option('--api-gateway-extensions', required=True, type=click.Path(exists=True),
              help='to add the the specification')
@click.option('--add-new-version', is_flag=True, default=False,
              help='of the RestAPI resource and replace all references')
@click.option('--keep', default=1, help='number of versions to keep, if --add-new-version is specified')
@click.argument('path', nargs=-1, required=True, type=click.Path(exists=True))
@click.pass_context
def swagger_document(ctx, resource, open_api_specification, api_gateway_extensions, path, add_new_version, keep):
    updater = RestAPIBodyUpdater()
    updater.main(resource, open_api_specification, api_gateway_extensions, list(path), add_new_version, keep,
                 ctx.obj['dry_run'], ctx.obj['verbose'])


@cli.command(name='lambda-inline-code', help=LambdaInlineCodeUpdater.__doc__)
@click.option('--resource', required=True, help='name of the AWS::Lambda::Function to update')
@click.option('--file', required=True, type=click.Path(exists=True), help='containing the source')
@click.argument('path', nargs=-1, required=True, type=click.Path(exists=True))
@click.pass_context
def lambda_body(ctx, resource, file, path):
    updater = LambdaInlineCodeUpdater()

    with open(file, 'r') as f:
        body = f.read()
    updater.main(resource, body, list(path), ctx.obj['dry_run'], ctx.obj['verbose'])


@cli.command(name='config-rule-inline-code', help=ConfigRuleInlineCodeUpdater.__doc__)
@click.option('--resource', required=True, help='name of the AWS::Config::ConfigRule')
@click.option('--file', required=True, type=click.Path(exists=True), help='containing the source')
@click.argument('path', nargs=-1, required=True, type=click.Path(exists=True))
@click.pass_context
def config_rule_body(ctx, resource, file, path):
    updater = ConfigRuleInlineCodeUpdater()

    with open(file, 'r') as f:
        body = f.read()
    updater.main(resource, body, list(path), ctx.obj['dry_run'], ctx.obj['verbose'])


cli.add_command(update_state_machine_definition)
cli.add_command(remove_resource)
cli.add_command(add_new_resources)
cli.add_command(packer_latest_ami)
cli.add_command(update_oidc_provider_thumbprint)

def main():
    cli()


if __name__ == '__main__':
    main()



import fnmatch
import json
import re

import boto3
import click
import sys


class PackerAMIUpdater(object):
    """
    Updates a packer.json source_ami_filter to the latest AMI version.
    With an ami-name-pattern of `Windows_Server-2016-English-Full-Base-*` it will
    update the following from:
\b
      "source_ami_filter": {
        "filters": {
          "virtualization-type": "hvm",
          "name": ""Windows_Server-2016-English-Full-Base-2020.01.15",
          "root-device-type": "ebs"
        },
        "owners": [ "801119661308" ],
      },
\b
    to:
\b
      "source_ami_filter": {
        "filters": {
          "virtualization-type": "hvm",
          "name": "Windows_Server-2016-English-Full-Base-2020.02.12",
          "root-device-type": "ebs"l
        },
        "owners": [ "801119661308" ],
      },
    """

    def __init__(self):
        self.packer: dict = None
        self.filename: str = None
        self.dirty: bool = False
        self.verbose: bool = False
        self.dry_run: bool = False
        self.ami_name_pattern: str = None
        self.ec2 = boto3.client("ec2")

    def is_source_filter_name_match(self, source_ami_filter: dict) -> bool:
        name = source_ami_filter.get("filters", {}).get("name")
        pattern = re.compile(fnmatch.translate(self.ami_name_pattern))
        return pattern.match(name) if name else None

    def create_describe_image_request(self, source_ami_filter: dict):
        result = {}
        filters = [
            {"Name": "name", "Values": [self.ami_name_pattern]},
            {"Name": "state", "Values": ["available"]},
        ]
        for k, v in filter(
            lambda k: k[0] not in ["name", "state"],
            source_ami_filter.get("filters", {}).items(),
        ):
            filters.append({"Name": k, "Values": v if isinstance(v, list) else [v]})
        result["Filters"] = filters

        owners = source_ami_filter.get("owners")
        if owners:
            result["Owners"] = owners

        return result

    def load(self):
        self.dirty = False
        with open(self.filename, "r") as f:
            self.packer = json.load(f)

    def save(self):
        if self.dry_run:
            return
        if self.dirty:
            with open(self.filename, "w") as f:
                self.packer = json.dump(self.packer, f, indent=2)

    def update(self):
        builders = filter(
            lambda b: b["type"]
            in [
                "amazon-chroot",
                "amazon-ebs",
                "amazon-ebssurrogate",
                "amazon-instance",
            ],
            self.packer.get("builders", []),
        )
        for builder in builders:
            source_ami_filter = builder.get("source_ami_filter", {})
            if self.is_source_filter_name_match(source_ami_filter):
                old_name = source_ami_filter.get("filters", {}).get("name")
                request = self.create_describe_image_request(source_ami_filter)
                images = sorted(
                    self.ec2.describe_images(**request)["Images"],
                    key=lambda c: c["CreationDate"],
                )
                if images:
                    name = images[-1]["Name"]
                    if name != old_name:
                        source_ami_filter["filters"]["name"] = name
                        self.dirty = True
                        sys.stderr.write(
                            f"INFO: replacing source ami filter image name with {name} of builder {builder['type']}"
                        )
                    else:
                        sys.stderr.write(
                            f"INFO: source ami filter already up to date with image name {name} of builder {builder['type']}"
                        )


@click.command(name="packer-latest-ami", help=PackerAMIUpdater.__doc__)
@click.option(
    "--ami-name-pattern",
    required=True,
    help="glob style pattern of the AMI image name to use",
)
@click.argument("path", nargs=-1, required=True, type=click.Path(exists=True))
@click.pass_context
def main(ctx, ami_name_pattern, path):
    updater = PackerAMIUpdater()
    updater.dry_run = ctx.obj.get("dry_run") if ctx.obj else False
    updater.verbose = ctx.obj.get("verbose") if ctx.obj else False
    updater.ami_name_pattern = ami_name_pattern
    for filename in path:
        updater.filename = filename
        updater.load()
        updater.update()
        updater.save()


if __name__ == "__main__":
    main()



import botocore
import datetime
from dateutil.tz import tzutc, tzlocal


request = {
    "Filters": [
        {"Name": "name", "Values": ["Windows_Server-2016-English-Full-Base*"]},
        {"Name": "state", "Values": ["available"]},
        {"Name": "virtualization-type", "Values": ["hvm"]},
        {"Name": "root-device-type", "Values": ["ebs"]},
    ],
    "Owners": ["801119661308"],
}
response = {
    "Images": [
        {
            "Architecture": "x86_64",
            "CreationDate": "2022-09-14T18:53:28.000Z",
            "ImageId": "ami-0d030edc442a8c188",
            "ImageLocation": "amazon/Windows_Server-2016-English-Full-Base-2022.09.14",
            "ImageType": "machine",
            "Public": True,
            "OwnerId": "801119661308",
            "Platform": "windows",
            "PlatformDetails": "Windows",
            "UsageOperation": "RunInstances:0002",
            "State": "available",
            "BlockDeviceMappings": [
                {
                    "DeviceName": "/dev/sda1",
                    "Ebs": {
                        "DeleteOnTermination": True,
                        "SnapshotId": "snap-0f7674948128a2988",
                        "VolumeSize": 30,
                        "VolumeType": "gp2",
                        "Encrypted": False,
                    },
                },
                {"DeviceName": "xvdca", "VirtualName": "ephemeral0"},
                {"DeviceName": "xvdcb", "VirtualName": "ephemeral1"},
                {"DeviceName": "xvdcc", "VirtualName": "ephemeral2"},
                {"DeviceName": "xvdcd", "VirtualName": "ephemeral3"},
                {"DeviceName": "xvdce", "VirtualName": "ephemeral4"},
                {"DeviceName": "xvdcf", "VirtualName": "ephemeral5"},
                {"DeviceName": "xvdcg", "VirtualName": "ephemeral6"},
                {"DeviceName": "xvdch", "VirtualName": "ephemeral7"},
                {"DeviceName": "xvdci", "VirtualName": "ephemeral8"},
                {"DeviceName": "xvdcj", "VirtualName": "ephemeral9"},
                {"DeviceName": "xvdck", "VirtualName": "ephemeral10"},
                {"DeviceName": "xvdcl", "VirtualName": "ephemeral11"},
                {"DeviceName": "xvdcm", "VirtualName": "ephemeral12"},
                {"DeviceName": "xvdcn", "VirtualName": "ephemeral13"},
                {"DeviceName": "xvdco", "VirtualName": "ephemeral14"},
                {"DeviceName": "xvdcp", "VirtualName": "ephemeral15"},
                {"DeviceName": "xvdcq", "VirtualName": "ephemeral16"},
                {"DeviceName": "xvdcr", "VirtualName": "ephemeral17"},
                {"DeviceName": "xvdcs", "VirtualName": "ephemeral18"},
                {"DeviceName": "xvdct", "VirtualName": "ephemeral19"},
                {"DeviceName": "xvdcu", "VirtualName": "ephemeral20"},
                {"DeviceName": "xvdcv", "VirtualName": "ephemeral21"},
                {"DeviceName": "xvdcw", "VirtualName": "ephemeral22"},
                {"DeviceName": "xvdcx", "VirtualName": "ephemeral23"},
                {"DeviceName": "xvdcy", "VirtualName": "ephemeral24"},
                {"DeviceName": "xvdcz", "VirtualName": "ephemeral25"},
            ],
            "Description": "Microsoft Windows Server 2016 with Desktop Experience Locale English AMI provided by Amazon",
            "EnaSupport": True,
            "Hypervisor": "xen",
            "ImageOwnerAlias": "amazon",
            "Name": "Windows_Server-2016-English-Full-Base-2022.09.14",
            "RootDeviceName": "/dev/sda1",
            "RootDeviceType": "ebs",
            "SriovNetSupport": "simple",
            "VirtualizationType": "hvm",
            "DeprecationTime": "2024-09-14T18:53:28.000Z",
        },
        {
            "Architecture": "x86_64",
            "CreationDate": "2022-08-10T07:21:09.000Z",
            "ImageId": "ami-0a27d685ce855d752",
            "ImageLocation": "amazon/Windows_Server-2016-English-Full-Base-2022.08.10",
            "ImageType": "machine",
            "Public": True,
            "OwnerId": "801119661308",
            "Platform": "windows",
            "PlatformDetails": "Windows",
            "UsageOperation": "RunInstances:0002",
            "State": "available",
            "BlockDeviceMappings": [
                {
                    "DeviceName": "/dev/sda1",
                    "Ebs": {
                        "DeleteOnTermination": True,
                        "SnapshotId": "snap-0ffc356891b0aefa9",
                        "VolumeSize": 30,
                        "VolumeType": "gp2",
                        "Encrypted": False,
                    },
                },
                {"DeviceName": "xvdca", "VirtualName": "ephemeral0"},
                {"DeviceName": "xvdcb", "VirtualName": "ephemeral1"},
                {"DeviceName": "xvdcc", "VirtualName": "ephemeral2"},
                {"DeviceName": "xvdcd", "VirtualName": "ephemeral3"},
                {"DeviceName": "xvdce", "VirtualName": "ephemeral4"},
                {"DeviceName": "xvdcf", "VirtualName": "ephemeral5"},
                {"DeviceName": "xvdcg", "VirtualName": "ephemeral6"},
                {"DeviceName": "xvdch", "VirtualName": "ephemeral7"},
                {"DeviceName": "xvdci", "VirtualName": "ephemeral8"},
                {"DeviceName": "xvdcj", "VirtualName": "ephemeral9"},
                {"DeviceName": "xvdck", "VirtualName": "ephemeral10"},
                {"DeviceName": "xvdcl", "VirtualName": "ephemeral11"},
                {"DeviceName": "xvdcm", "VirtualName": "ephemeral12"},
                {"DeviceName": "xvdcn", "VirtualName": "ephemeral13"},
                {"DeviceName": "xvdco", "VirtualName": "ephemeral14"},
                {"DeviceName": "xvdcp", "VirtualName": "ephemeral15"},
                {"DeviceName": "xvdcq", "VirtualName": "ephemeral16"},
                {"DeviceName": "xvdcr", "VirtualName": "ephemeral17"},
                {"DeviceName": "xvdcs", "VirtualName": "ephemeral18"},
                {"DeviceName": "xvdct", "VirtualName": "ephemeral19"},
                {"DeviceName": "xvdcu", "VirtualName": "ephemeral20"},
                {"DeviceName": "xvdcv", "VirtualName": "ephemeral21"},
                {"DeviceName": "xvdcw", "VirtualName": "ephemeral22"},
                {"DeviceName": "xvdcx", "VirtualName": "ephemeral23"},
                {"DeviceName": "xvdcy", "VirtualName": "ephemeral24"},
                {"DeviceName": "xvdcz", "VirtualName": "ephemeral25"},
            ],
            "Description": "Microsoft Windows Server 2016 with Desktop Experience Locale English AMI provided by Amazon",
            "EnaSupport": True,
            "Hypervisor": "xen",
            "ImageOwnerAlias": "amazon",
            "Name": "Windows_Server-2016-English-Full-Base-2022.08.10",
            "RootDeviceName": "/dev/sda1",
            "RootDeviceType": "ebs",
            "SriovNetSupport": "simple",
            "VirtualizationType": "hvm",
            "DeprecationTime": "2024-08-10T07:21:09.000Z",
        },
        {
            "Architecture": "x86_64",
            "CreationDate": "2022-10-28T01:24:13.000Z",
            "ImageId": "ami-01527f1bfc6f63543",
            "ImageLocation": "amazon/Windows_Server-2016-English-Full-Base-2022.10.27",
            "ImageType": "machine",
            "Public": True,
            "OwnerId": "801119661308",
            "Platform": "windows",
            "PlatformDetails": "Windows",
            "UsageOperation": "RunInstances:0002",
            "State": "available",
            "BlockDeviceMappings": [
                {
                    "DeviceName": "/dev/sda1",
                    "Ebs": {
                        "DeleteOnTermination": True,
                        "SnapshotId": "snap-0636cccace06ba03d",
                        "VolumeSize": 30,
                        "VolumeType": "gp2",
                        "Encrypted": False,
                    },
                },
                {"DeviceName": "xvdca", "VirtualName": "ephemeral0"},
                {"DeviceName": "xvdcb", "VirtualName": "ephemeral1"},
                {"DeviceName": "xvdcc", "VirtualName": "ephemeral2"},
                {"DeviceName": "xvdcd", "VirtualName": "ephemeral3"},
                {"DeviceName": "xvdce", "VirtualName": "ephemeral4"},
                {"DeviceName": "xvdcf", "VirtualName": "ephemeral5"},
                {"DeviceName": "xvdcg", "VirtualName": "ephemeral6"},
                {"DeviceName": "xvdch", "VirtualName": "ephemeral7"},
                {"DeviceName": "xvdci", "VirtualName": "ephemeral8"},
                {"DeviceName": "xvdcj", "VirtualName": "ephemeral9"},
                {"DeviceName": "xvdck", "VirtualName": "ephemeral10"},
                {"DeviceName": "xvdcl", "VirtualName": "ephemeral11"},
                {"DeviceName": "xvdcm", "VirtualName": "ephemeral12"},
                {"DeviceName": "xvdcn", "VirtualName": "ephemeral13"},
                {"DeviceName": "xvdco", "VirtualName": "ephemeral14"},
                {"DeviceName": "xvdcp", "VirtualName": "ephemeral15"},
                {"DeviceName": "xvdcq", "VirtualName": "ephemeral16"},
                {"DeviceName": "xvdcr", "VirtualName": "ephemeral17"},
                {"DeviceName": "xvdcs", "VirtualName": "ephemeral18"},
                {"DeviceName": "xvdct", "VirtualName": "ephemeral19"},
                {"DeviceName": "xvdcu", "VirtualName": "ephemeral20"},
                {"DeviceName": "xvdcv", "VirtualName": "ephemeral21"},
                {"DeviceName": "xvdcw", "VirtualName": "ephemeral22"},
                {"DeviceName": "xvdcx", "VirtualName": "ephemeral23"},
                {"DeviceName": "xvdcy", "VirtualName": "ephemeral24"},
                {"DeviceName": "xvdcz", "VirtualName": "ephemeral25"},
            ],
            "Description": "Microsoft Windows Server 2016 with Desktop Experience Locale English AMI provided by Amazon",
            "EnaSupport": True,
            "Hypervisor": "xen",
            "ImageOwnerAlias": "amazon",
            "Name": "Windows_Server-2016-English-Full-Base-2022.10.27",
            "RootDeviceName": "/dev/sda1",
            "RootDeviceType": "ebs",
            "SriovNetSupport": "simple",
            "VirtualizationType": "hvm",
            "DeprecationTime": "2024-10-28T01:24:13.000Z",
        },
        {
            "Architecture": "x86_64",
            "CreationDate": "2022-11-10T13:31:43.000Z",
            "ImageId": "ami-0dd4486f71af814de",
            "ImageLocation": "amazon/Windows_Server-2016-English-Full-Base-2022.11.10",
            "ImageType": "machine",
            "Public": True,
            "OwnerId": "801119661308",
            "Platform": "windows",
            "PlatformDetails": "Windows",
            "UsageOperation": "RunInstances:0002",
            "State": "available",
            "BlockDeviceMappings": [
                {
                    "DeviceName": "/dev/sda1",
                    "Ebs": {
                        "DeleteOnTermination": True,
                        "SnapshotId": "snap-099d325d96bf48c86",
                        "VolumeSize": 30,
                        "VolumeType": "gp2",
                        "Encrypted": False,
                    },
                },
                {"DeviceName": "xvdca", "VirtualName": "ephemeral0"},
                {"DeviceName": "xvdcb", "VirtualName": "ephemeral1"},
                {"DeviceName": "xvdcc", "VirtualName": "ephemeral2"},
                {"DeviceName": "xvdcd", "VirtualName": "ephemeral3"},
                {"DeviceName": "xvdce", "VirtualName": "ephemeral4"},
                {"DeviceName": "xvdcf", "VirtualName": "ephemeral5"},
                {"DeviceName": "xvdcg", "VirtualName": "ephemeral6"},
                {"DeviceName": "xvdch", "VirtualName": "ephemeral7"},
                {"DeviceName": "xvdci", "VirtualName": "ephemeral8"},
                {"DeviceName": "xvdcj", "VirtualName": "ephemeral9"},
                {"DeviceName": "xvdck", "VirtualName": "ephemeral10"},
                {"DeviceName": "xvdcl", "VirtualName": "ephemeral11"},
                {"DeviceName": "xvdcm", "VirtualName": "ephemeral12"},
                {"DeviceName": "xvdcn", "VirtualName": "ephemeral13"},
                {"DeviceName": "xvdco", "VirtualName": "ephemeral14"},
                {"DeviceName": "xvdcp", "VirtualName": "ephemeral15"},
                {"DeviceName": "xvdcq", "VirtualName": "ephemeral16"},
                {"DeviceName": "xvdcr", "VirtualName": "ephemeral17"},
                {"DeviceName": "xvdcs", "VirtualName": "ephemeral18"},
                {"DeviceName": "xvdct", "VirtualName": "ephemeral19"},
                {"DeviceName": "xvdcu", "VirtualName": "ephemeral20"},
                {"DeviceName": "xvdcv", "VirtualName": "ephemeral21"},
                {"DeviceName": "xvdcw", "VirtualName": "ephemeral22"},
                {"DeviceName": "xvdcx", "VirtualName": "ephemeral23"},
                {"DeviceName": "xvdcy", "VirtualName": "ephemeral24"},
                {"DeviceName": "xvdcz", "VirtualName": "ephemeral25"},
            ],
            "Description": "Microsoft Windows Server 2016 with Desktop Experience Locale English AMI provided by Amazon",
            "EnaSupport": True,
            "Hypervisor": "xen",
            "ImageOwnerAlias": "amazon",
            "Name": "Windows_Server-2016-English-Full-Base-2022.11.10",
            "RootDeviceName": "/dev/sda1",
            "RootDeviceType": "ebs",
            "SriovNetSupport": "simple",
            "VirtualizationType": "hvm",
            "DeprecationTime": "2024-11-10T13:31:43.000Z",
        },
        {
            "Architecture": "x86_64",
            "CreationDate": "2022-10-12T16:06:39.000Z",
            "ImageId": "ami-07cd9d4167c228ea0",
            "ImageLocation": "amazon/Windows_Server-2016-English-Full-Base-2022.10.12",
            "ImageType": "machine",
            "Public": True,
            "OwnerId": "801119661308",
            "Platform": "windows",
            "PlatformDetails": "Windows",
            "UsageOperation": "RunInstances:0002",
            "State": "available",
            "BlockDeviceMappings": [
                {
                    "DeviceName": "/dev/sda1",
                    "Ebs": {
                        "DeleteOnTermination": True,
                        "SnapshotId": "snap-0f492bcf1abc69555",
                        "VolumeSize": 30,
                        "VolumeType": "gp2",
                        "Encrypted": False,
                    },
                },
                {"DeviceName": "xvdca", "VirtualName": "ephemeral0"},
                {"DeviceName": "xvdcb", "VirtualName": "ephemeral1"},
                {"DeviceName": "xvdcc", "VirtualName": "ephemeral2"},
                {"DeviceName": "xvdcd", "VirtualName": "ephemeral3"},
                {"DeviceName": "xvdce", "VirtualName": "ephemeral4"},
                {"DeviceName": "xvdcf", "VirtualName": "ephemeral5"},
                {"DeviceName": "xvdcg", "VirtualName": "ephemeral6"},
                {"DeviceName": "xvdch", "VirtualName": "ephemeral7"},
                {"DeviceName": "xvdci", "VirtualName": "ephemeral8"},
                {"DeviceName": "xvdcj", "VirtualName": "ephemeral9"},
                {"DeviceName": "xvdck", "VirtualName": "ephemeral10"},
                {"DeviceName": "xvdcl", "VirtualName": "ephemeral11"},
                {"DeviceName": "xvdcm", "VirtualName": "ephemeral12"},
                {"DeviceName": "xvdcn", "VirtualName": "ephemeral13"},
                {"DeviceName": "xvdco", "VirtualName": "ephemeral14"},
                {"DeviceName": "xvdcp", "VirtualName": "ephemeral15"},
                {"DeviceName": "xvdcq", "VirtualName": "ephemeral16"},
                {"DeviceName": "xvdcr", "VirtualName": "ephemeral17"},
                {"DeviceName": "xvdcs", "VirtualName": "ephemeral18"},
                {"DeviceName": "xvdct", "VirtualName": "ephemeral19"},
                {"DeviceName": "xvdcu", "VirtualName": "ephemeral20"},
                {"DeviceName": "xvdcv", "VirtualName": "ephemeral21"},
                {"DeviceName": "xvdcw", "VirtualName": "ephemeral22"},
                {"DeviceName": "xvdcx", "VirtualName": "ephemeral23"},
                {"DeviceName": "xvdcy", "VirtualName": "ephemeral24"},
                {"DeviceName": "xvdcz", "VirtualName": "ephemeral25"},
            ],
            "Description": "Microsoft Windows Server 2016 with Desktop Experience Locale English AMI provided by Amazon",
            "EnaSupport": True,
            "Hypervisor": "xen",
            "ImageOwnerAlias": "amazon",
            "Name": "Windows_Server-2016-English-Full-Base-2022.10.12",
            "RootDeviceName": "/dev/sda1",
            "RootDeviceType": "ebs",
            "SriovNetSupport": "simple",
            "VirtualizationType": "hvm",
            "DeprecationTime": "2024-10-12T16:06:39.000Z",
        },
    ],
    "ResponseMetadata": {
        "RequestId": "595905ff-d097-4e65-8aad-82e3ff3bf626",
        "HTTPStatusCode": 200,
        "HTTPHeaders": {
            "x-amzn-requestid": "595905ff-d097-4e65-8aad-82e3ff3bf626",
            "cache-control": "no-cache, no-store",
            "strict-transport-security": "max-age=31536000; includeSubDomains",
            "vary": "accept-encoding",
            "content-type": "text/xml;charset=UTF-8",
            "transfer-encoding": "chunked",
            "date": "Sat, 12 Nov 2022 11:10:04 GMT",
            "server": "AmazonEC2",
        },
        "RetryAttempts": 0,
    },
}



#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
#   Copyright 2018 binx.io B.V.
import sys
from .cfn_updater import CfnUpdater
from ruamel.yaml.scalarstring import PreservedScalarString


class ConfigRuleInlineCodeUpdater(CfnUpdater):
    """
    Updates the inline code of an AWS::Config::ConfigRule resource.

\b
  ConfigRule:
    Type: AWS::Config::ConfigRule

\b
  ConfigRule:
    Type: AWS::Config::ConfigRule
    Properties:
      Source:
        Owner: CUSTOM_POLICY
        CustomPolicyDetails:
          EnableDebugLogDelivery: true
          PolicyRuntime: guard-2.x.x
          PolicyText: |-
            rule name when resourceType == "AWS::S3::Bucket" {
                ...
            }
          ...
    """

    def __init__(self):
        super(ConfigRuleInlineCodeUpdater, self).__init__()
        self._image = []
        self.resource_name = None
        self.code = None

    def update_template(self):
        """
        updates the Code property of a AWS::Config::ConfigRule resource of name `self.resource` to `self.code`
        """
        resource = self.template.get('Resources', {}).get(self.resource_name, None)
        if (
                resource
                and resource["Type"] == "AWS::Config::ConfigRule"
                and resource.get("Owner", "CUSTOM_POLICY") == "CUSTOM_POLICY"
        ):
            properties = resource.get('Properties', {})
            source = properties.get('Source', {})
            details = source.get('CustomPolicyDetails', {})
            old_code = details.get('PolicyText', None)

            if old_code != self.code:
                sys.stderr.write(
                    'INFO: updating policy text of config rule {} in {}\n'.format(self.resource_name, self.filename))
                if 'Properties' not in resource:
                    resource['Properties'] = {}
                if 'Owner' not in resource['Properties']:
                    resource['Properties']['Owner'] = 'CUSTOM_POLICY'
                if 'Source' not in resource['Properties']:
                    resource['Properties']['Source'] = {}
                if 'CustomPolicyDetails' not in resource['Properties']['Source']:
                    resource['Properties']['Source']['CustomPolicyDetails'] = {}

                resource['Properties']['Source']['CustomPolicyDetails']['PolicyText'] = PreservedScalarString(self.code)
                self.dirty = True
        elif resource:
            sys.stderr.write(
                'WARN: resource {} in {} is not a CUSTOM_POLICY of type AWS::Config::ConfigRule\n'.format(self.resource_name, self.filename))

    def main(self, resource, code, paths, dry_run, verbose):
        self.resource_name = resource
        self.code = code
        self.dry_run = dry_run
        self.verbose = verbose
        self.update(paths)



"""
Programmatically update CloudFormation templates
"""
from os import path
from setuptools import find_packages, setup

dependencies = ['click', 'boto3', 'cfn-flip', 'pytz', 'pytest', 'croniter', 'tzlocal', 'click-datetime', 'ruamel.yaml', 'requests', 'jsonmerge', 'cryptography']

with open(path.join(path.abspath(path.dirname(__file__)), 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='aws-cfn-update',
    version="0.7.2",
    url='https://github.com/binxio/aws-cfn-update',
    license='BSD',
    author='Mark van Holsteijn',
    author_email='mark@binx.io',
    description='Programmatically update CloudFormation templates',
    long_description=long_description,
    long_description_content_type='text/markdown',
    packages=find_packages(exclude=['tests']),
    include_package_data=True,
    zip_safe=False,
    platforms='any',
    install_requires=dependencies,
    setup_requires=['pytest-runner'],
    tests_require=dependencies + ['pytest'],
    test_suite='tests',
    entry_points={
        'console_scripts': [
            'aws-cfn-update = aws_cfn_update.cli:main',
        ],
    },
    classifiers=[
        # As from http://pypi.python.org/pypi?%3Aaction=list_classifiers
        # 'Development Status :: 1 - Planning',
        # 'Development Status :: 2 - Pre-Alpha',
        # 'Development Status :: 3 - Alpha',
        # 'Development Status :: 4 - Beta',
        'Development Status :: 5 - Production/Stable',
        # 'Development Status :: 6 - Mature',
        # 'Development Status :: 7 - Inactive',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: POSIX',
        'Operating System :: MacOS',
        'Operating System :: Unix',
        'Operating System :: Microsoft :: Windows',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ]
)



#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
#   Copyright 2018 binx.io B.V.
import difflib
import re
import sys
from io import BytesIO
from io import StringIO

import jsonmerge
from ruamel.yaml import YAML

from aws_cfn_update.cfn_updater import CfnUpdater
from aws_cfn_update.replace_references import replace_references


class RestAPIBodyUpdater(CfnUpdater):
    """
    Updates the body of a REST API Resource, with an standard Open API
    specification merged with AWS API Gateway extensions.

    If you specify --add-new-version, it will create a new
    version of the resource and update all references
    to it. This will enforce the deployment of the new api.

    If you want to keep the previous definition, specify --keep to
    a value of 2 or higher. This might be handy if you have old clients
    still accessing the old version of the API.

    If no changes are detected, no changes are made. Please make
    sure that all dictionary keys in th specifications are strings,
    not integers (especially the case with `responses`).
    When updating json CFN templates, the compare algorithm does
    not work properly.
    """

    def __init__(self):
        super(RestAPIBodyUpdater, self).__init__()
        self.resource_name = None
        self.open_api_specification = None
        self.api_gateway_extensions = None
        self.template = None
        self.add_new_version = False
        self.verbose = False
        self.dry_run = False
        self.keep = 1
        self.yaml = YAML()
        self._body = {}

    def load_and_merge_swagger_body(self):
        with open(self.open_api_specification, 'r') as f:
            body = self.yaml.load(f)
        with open(self.api_gateway_extensions, 'r') as f:
            extensions = self.yaml.load(f)

        self.body = jsonmerge.merge(body, extensions)

    @property
    def body(self):
        return self._body

    @body.setter
    def body(self, body):
        self._body = body
        self.body_as_string = self.yaml_dump_to_str(self.body) if body else ''

    def yaml_dump_to_str(self, dict):
        s = StringIO()
        self.yaml.dump(dict, s)
        return s.getvalue()

    def resource_name_pattern(self):
        return re.compile('^(?P<basename>{})(v(?P<version>[0-9]+))?$'.format(self.resource_name))

    def find_matching_resources(self):
        """
        find all resources matching the pattern `<resource_name>v[0-9]+`, sort listed
        """
        pattern = self.resource_name_pattern()

        result = []
        resources = self.template['Resources'] if 'Resources' in self.template else None
        if resources:
            result = list(filter(lambda name: 'Type' in resources[name] and resources[name][
                'Type'] == 'AWS::ApiGateway::RestApi', filter(lambda name: pattern.match(name), resources)))
            result.sort(
                key=lambda n: int(pattern.match(n).group('version')) if pattern.match(n).group('version') else -1)
        return result

    def new_resource_name(self, name):
        match = self.resource_name_pattern().match(name)
        version = int(match.group('version')) if match.group('version') else 0
        return '{}v{}'.format(match.group('basename'), version + 1)

    def copy_resource(self, resource):
        """
        Copy the resource using dump and reload. dictionary deepcopy() did not work on yaml dictionaries.
        """
        bytes = BytesIO()
        self.yaml.dump(resource, bytes)
        bytes.seek(0)
        result = self.yaml.load(bytes)
        return result

    def update_template(self):
        resources = self.find_matching_resources()
        if not resources:
            return

        name = resources[-1]
        rest_api_gateway = self.template['Resources'][name]
        current_body = rest_api_gateway.get('Properties', {}).get('Body', {})

        current_body_as_string = self.yaml_dump_to_str(current_body) if current_body else ''
        if self.body_as_string != current_body_as_string:

            if self.verbose:
                for text in difflib.unified_diff(current_body_as_string.split("\n"), self.body_as_string.split("\n")):
                    if text[:3] not in ['---', '+++']:
                        sys.stderr.write('{}\n'.format(text))

            rest_api_gateway = self.copy_resource(rest_api_gateway)
            if not 'Properties' in rest_api_gateway:
                rest_api_gateway['Properties'] = {}

            rest_api_gateway['Properties']['Body'] = self.body

            if self.add_new_version:
                new_name = self.new_resource_name(name)
                sys.stderr.write(
                    'INFO: adding resource {} with new swagger body in template {}\n'.format(new_name, self.filename))
            else:
                new_name = name
                sys.stderr.write(
                    'INFO: updating resource {} with swagger body in template {}\n'.format(new_name, self.filename))

            if self.add_new_version:
                replace_references(self.template, name, new_name)
                for i in range(0, len(resources) - (self.keep - 1)):
                    sys.stderr.write(
                        'INFO: removing resource {} from template {}\n'.format(resources[i], self.filename))
                    del self.template['Resources'][resources[i]]
            self.template['Resources'][new_name] = rest_api_gateway

            self.dirty = True
        else:
            sys.stderr.write(
                'INFO: no changes of swagger body of {} in template {}\n'.format(self.resource_name, self.filename))

    def main(self, resource_name, open_api_specification, api_gateway_extensions, path, add_new_version, keep, dry_run,
             verbose):
        self.dry_run = dry_run
        self.verbose = verbose
        self.resource_name = resource_name
        self.open_api_specification = open_api_specification
        self.api_gateway_extensions = api_gateway_extensions
        self.add_new_version = add_new_version
        self.keep = keep if keep > 0 else 1
        self.load_and_merge_swagger_body()
        self.update(path)



#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
#   Copyright 2018 binx.io B.V.
import sys

import click
from ruamel.yaml.scalarstring import LiteralScalarString

from aws_cfn_update.cfn_updater import CfnUpdater


class StateMachineDefinitionUpdater(CfnUpdater):
    """
    Updates the definition of an AWS::StepFunctions::StateMachine.

    The definition is read from the file specified by --definition.
    By default, the content will be passed into the Fn::Sub function to
    allow references to parameters and resource attributes in the template.

    If you do not want substitution for your definition, specify --no-fn-sub.

    """
    def __init__(self):
        super(StateMachineDefinitionUpdater, self).__init__()
        self.resource_name = None
        self._definition = None
        self.definition_file = None
        self.template = None
        self.with_fn_sub = True
        self.verbose = False
        self.dry_run = False

    def new_value(self, definition):
        if self.with_fn_sub:
            return isinstance(definition, dict) and definition.get('Fn::Sub') == self.definition, {'Fn::Sub': self.definition}
        else:
            return definition == self.definition, self.definition

    def update_template(self):
        """
        updates the Definition of Step Function StateMachine.
        """
        resources = self.template.get('Resources', {})
        for name, resource in map(lambda n: (n, resources[n]), filter(lambda n: n == self.resource_name, resources)):
            if 'Properties' not in resource:
                resource['Properties'] = {}

            same, new_definition = self.new_value(resource['Properties'].get('DefinitionString'))
            if not same:
                sys.stderr.write('INFO: updating definition of state machine {}\n'.format(self.resource_name))
                resource['Properties']['DefinitionString'] = new_definition
                self.dirty = True
            else:
                if self.verbose:
                    sys.stderr.write(
                        'INFO: no changes to definition of state machine {}\n'.format(self.resource_name))

    @property
    def definition(self):
        return self._definition

    @definition.setter
    def definition(self, definition):
        self._definition = LiteralScalarString(definition) if not isinstance(definition, LiteralScalarString) else definition

    def read_definition(self, filename):
        self.definition_file = filename
        with open(self.definition_file, 'r') as f:
            self.definition = LiteralScalarString(f.read())

    def main(self, resource_name, definition_file, with_fn_sub, path, dry_run, verbose):
        self.dry_run = dry_run
        self.verbose = verbose
        self.resource_name = resource_name
        self.with_fn_sub = with_fn_sub
        self.read_definition(definition_file)
        self.update(path)


@click.command(name='state-machine-definition', help=StateMachineDefinitionUpdater.__doc__)
@click.option('--resource', required=True, help='AWS::StepFunctions::StateMachine definition to update')
@click.option('--definition', required=True, type=click.Path(exists=True), help='of the state machine')
@click.option('--fn-sub/--no-fn-sub', required=False, default=True, help='for the definition')
@click.argument('path', nargs=-1, required=True, type=click.Path(exists=True))
@click.pass_context
def update_state_machine_definition(ctx, resource, definition, fn_sub, path):
    updater = StateMachineDefinitionUpdater()
    updater.main(resource, definition, fn_sub, list(path), ctx.obj['dry_run'], ctx.obj['verbose'])
